use core::f64;

use crate::{errors::RoctreeError, primitives::Box3, stl::parser::StlSolid};

/// Number of potential split planes to check
const NBINS: usize = 8;

/// Struct to hold bin information, this includes an axis-aligned bounding box as well as the count
/// of triangles with the
#[derive(Default, Clone)]
struct Bin {
    aabb: Box3,
    tri_count: usize,
}

/// A node representing a node single in the bounding volume hierarchy
#[derive(Default)]
struct BvhNode {
    aabb: Box3,
    left_child: usize,
    right_child: usize,
    prim_idx: usize,
    prim_count: usize,
}

impl BvhNode {
    /// This computes the computational cost associated with the specific node.
    /// The cost is an aggregate that considers the size of the bounding box (how likely it will
    /// intersect a ray) * the number of triangles (the number of ray-triangle checks we have to
    /// perform if an intersect occurs)
    fn calculate_node_cost(&self) -> f32 {
        self.aabb.surface_area() * self.prim_count as f32
    }
}

/// Struct representing the total bounding volume hierarchy
pub struct Bvh {
    nodes: Vec<BvhNode>,
    tri_idx: Vec<usize>,
}

/// Return type for get_subdivision_position_and_axis
struct BestSplitPlane {
    axis: usize,
    pos: f64,
    cost: f64,
}

impl Bvh {
    /// Create a new Bounding volume hierarchy from an STL solid.
    pub fn new(solid: &StlSolid) -> Result<Bvh, RoctreeError> {
        // Initialize the root node
        let mut nodes = vec![BvhNode {
            prim_idx: 0,
            prim_count: solid.triangles.len(),
            ..Default::default()
        }];
        nodes[0]
            .aabb
            .shrink_wrap_primitives(&solid.vertices, &solid.triangles);

        let mut bvh = Bvh {
            nodes,
            tri_idx: Vec::from_iter(0..solid.triangles.len()),
        };

        bvh.subdivide(0, solid)?;
        Ok(bvh)
    }

    /// Subdivide a node, subdivision doesn't occur if the node has only 2 primitives
    pub fn subdivide(&mut self, node_idx: usize, solid: &StlSolid) -> Result<(), RoctreeError> {
        let BestSplitPlane {
            cost,
            axis: split_axis,
            pos: split_position,
        } = self.get_best_split(node_idx, solid)?;

        // If the split cost is greater than keeping the node intact just return
        if cost > self.nodes[node_idx].calculate_node_cost() as f64 {
            return Ok(());
        }

        let prim_count = self.nodes[node_idx].prim_count;
        let prim_idx = self.nodes[node_idx].prim_idx;
        // Reorganize the triangles so the triangles are ordered, this allows triangles belonging
        // to a specific node to be adjacent
        let mut start_idx = prim_idx;
        let mut end_idx = prim_idx + prim_count - 1;
        while start_idx <= end_idx {
            if solid.triangle_centroids[self.tri_idx[start_idx]][split_axis] < split_position as f32
            {
                start_idx += 1;
            } else {
                self.tri_idx.swap(start_idx, end_idx);
                end_idx -= 1;
            }
        }

        // Now split the primitives into two nodes
        let left_count = start_idx - prim_idx;
        let right_count = prim_count - left_count;

        // If all primitives belong to either the left or right half exit because the parent node
        // is a leaf
        if left_count == 0 || left_count == prim_count {
            return Ok(());
        }

        // Otherwise split the nodes into left and right child
        let left_child = self.nodes.len();
        let right_child = self.nodes.len() + 1;

        // FIX: Having to copy the [usize; 3] that defines each triangle just to create the
        // axis-aligned bounding box seems wasteful although it makes the code interface a lot
        // cleaner. Maybe spend some time to redesign this part to make it cleaner once we really
        // start focusing on performance tuning
        let left_prim_triangles: Vec<_> = (prim_idx..left_count)
            .map(|idx| solid.triangles[idx])
            .collect();
        let mut aabb = Box3::default();
        aabb.shrink_wrap_primitives(&solid.vertices, &left_prim_triangles);

        self.nodes.push(BvhNode {
            aabb,
            left_child: 0,
            right_child: 0,
            prim_idx,
            prim_count: left_count,
        });

        let right_prim_triangles: Vec<_> = ((prim_idx + left_count)..prim_count)
            .map(|idx| solid.triangles[idx])
            .collect();
        let mut aabb = Box3::default();
        aabb.shrink_wrap_primitives(&solid.vertices, &right_prim_triangles);

        self.nodes.push(BvhNode {
            aabb,
            left_child: 0,
            right_child: 0,
            prim_idx: prim_idx + left_count,
            prim_count: right_count,
        });
        self.nodes[node_idx].prim_count = 0;
        self.nodes[node_idx].left_child = left_child;
        self.nodes[node_idx].right_child = right_child;

        // Recurse to divide children
        self.subdivide(left_child, solid)?;
        self.subdivide(right_child, solid)?;

        Ok(())
    }

    /// Get the best split plane based on the surface area heuristic
    /// This procedure works by binning
    fn get_best_split(
        &self,
        node_idx: usize,
        solid: &StlSolid,
    ) -> Result<BestSplitPlane, RoctreeError> {
        // Initialize the best split plane information with bad values so we know if no plane gets
        // assigned
        let mut best_cost = f64::INFINITY;
        let mut best_pos = f64::NAN;

        // Pick the longest axis as the splitting axis
        let axis = self.nodes[node_idx]
            .aabb
            .get_extents()
            .iter()
            .enumerate()
            .fold((0_usize, &f32::NEG_INFINITY), |acc, a| {
                if acc.1 > a.1 {
                    acc
                } else {
                    a
                }
            })
            .0;

        let node = &self.nodes[node_idx];
        // Get minimum and maximum bounds over the centroids off the triangles. This improves
        // the binning compared to doing the binning over the vertices
        let (min_bd, max_bd) = self.tri_idx[node.prim_idx..node.prim_idx + node.prim_count]
            .iter()
            .fold((f64::INFINITY, f64::NEG_INFINITY), |bounds, tri_idx| {
                (
                    bounds
                        .0
                        .min(solid.triangle_centroids[*tri_idx][axis].into()),
                    bounds
                        .1
                        .max(solid.triangle_centroids[*tri_idx][axis].into()),
                )
            });
        if min_bd == max_bd {
            return Err(RoctreeError::SplitPlaneError);
        };

        let mut bins = vec![Bin::default(); NBINS];
        let scale = NBINS as f64 / (max_bd - min_bd);

        // Assign the triangles to each of the bins
        for tri_idx in self.tri_idx[node.prim_idx..node.prim_idx + node.prim_count].iter() {
            let bin_index = usize::min(
                NBINS - 1,
                ((solid.triangle_centroids[*tri_idx][axis] as f64 - min_bd) * scale) as usize,
            );

            bins[bin_index].tri_count += 1;
            bins[bin_index]
                .aabb
                .grow(solid.vertices[solid.triangles[*tri_idx][0]]);
            bins[bin_index]
                .aabb
                .grow(solid.vertices[solid.triangles[*tri_idx][1]]);
            bins[bin_index]
                .aabb
                .grow(solid.vertices[solid.triangles[*tri_idx][2]]);
        }

        // Now get the left and right counts for each split plane (based on the bins)
        let mut left_area = [0.0; NBINS - 1];
        let mut right_area = [0.0; NBINS - 1];
        let mut left_count = [0; NBINS - 1];
        let mut right_count = [0; NBINS - 1];
        let mut left_box = Box3::default();
        let mut right_box = Box3::default();
        let mut left_sum = 0;
        let mut right_sum = 0;
        for plane_idx in 0..NBINS - 1 {
            // Calculate the left side of the plane
            left_sum += bins[plane_idx].tri_count;
            left_count[plane_idx] = left_sum;
            left_box.grow_by_other_box(&bins[plane_idx].aabb);
            left_area[plane_idx] = left_box.surface_area();

            // Calculate the right side of the plane
            right_sum += bins[NBINS - 1 - plane_idx].tri_count;
            right_count[NBINS - 2 - plane_idx] = right_sum;
            right_box.grow_by_other_box(&bins[NBINS - 1 - plane_idx].aabb);
            right_area[NBINS - 2 - plane_idx] = right_box.surface_area();
        }
        let scale = 1.0 / scale;
        for plane_idx in 0..NBINS - 1 {
            let plane_cost: f64 = (left_count[plane_idx] as f32 * left_area[plane_idx]
                + right_count[plane_idx] as f32 * right_area[plane_idx])
                .into();
            if plane_cost < best_cost {
                best_cost = plane_cost;
                best_pos = scale * plane_idx as f64;
            }
        }

        Ok(BestSplitPlane {
            axis,
            pos: best_pos,
            cost: best_cost,
        })
    }
}
